**************************************************************************;
** Macro name      : bspline                                            **;
** Version number  : 1.02                                               **;
** Authors         : Alberto Garcia-Hernandez                           **;
**                   Dimitris Rizopoulos                                **;
** Date            : 23-Jan-2015                                        **;
** SAS version     : 9.0 or higher                                      **;
** Purpose         : Calculate the Bspline Basis functions              **;
** Parameters      : indata       = Input Dataset                       **;
**                   outdata      = Output Dataset                      **;
**                   invariable   = Variable that stores T              **;
**                   outarray     = Array/Matrix to store the basis     **;
**                   dimout = outarray  is a matrix of dimension        **;
**                            Dimeout x Number of basis functions       **; 
**                            Use dimout=1 for the basis of variable T  **;
**                            Use dimout>1 when bases are required for a**;
**                            vector of times                           **;  
**                   posout = position of the basis in the output matrix**;
**                   lowerknot= lower boundary knot in hex16. format    **;
**                   upperknot = upper boundary knot  in hex16. format  **;
**                   Nknots    = number of internal knots (max=3)       **;
**                   degree     = polynomial degree                     **;
**                   knot1, knot2,... knot9 = internal knots value      **;
**                               in hex16. format     				    **;
**                   incderiv = 1 if derivatives of the functions are   **;
**                             required                                 **;
**                   incderiv2= YES if 2nd derivatives of the functions **;
**                             required                                 **;
**                   dif      = difference used to calculate derivatives**;
**                                                                      **;
**                        MACRO HISTORY                                 **;
** Version     Date        Changes                                      **;
** 1.01        15/12/2014  First Production Version                     **;
** 1.02        23/02/2015  New checks added for input knots not in      **;
**                         16-characters Hex16. format.                 **;
**                                                                      **;
** Copyright (C) 2014      Alberto Garcia Hernandez                     **;
**                         Dimitris Rizopoulos                          **;
** This program is free software: you can redistribute it and/or modify **;
** it under the terms of the GNU General Public License as published by **;
** the Free Software Foundation, either version 3 of the License, or    **;
** (at your option) any later version.                                  **;
**                                                                      **;
** This program is distributed in the hope that it will be useful,      **;
** but WITHOUT ANY WARRANTY, without even the implied warranty of       **;
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        **;
** GNU General Public License for more details.                         **;
**                                                                      **;
** You should have received a copy of the GNU General Public License    **;
** along with this program.  If not, see <http://www.gnu.org/licenses/  **;
** Everyone is permitted to copy and distribute verbatim copies         **;
** of this program, but changing it is not allowed.                     **;
**                                                                      **;
** Contact email: alberto.garciahernandez@astellas.com                  **;
**************************************************************************;

%macro bspline(indata=, outdata=, invariable=, outarray=, 
               dimout=1, posout=1, lowerknot=, upperknot=, Nknots=0, 
               degree=1, knot1=, knot2=, knot3=, knot4=, knot5=, knot6=, knot7=,
               knot8=,knot9=, incderiv=NO, incderiv2=NO, dif=0.01, nparam_var=longi_nparam, if=);


%let order=%eval(&degree.+1);              *Order m = n+1;
%let dim= %eval(&Nknots. + 2*&order.);     *dimension;
%let nparam=%eval(&order. + &Nknots.);     *number of parameters;

%if &nparam_var. NE %then %do; 
  %let &nparam_var. = &nparam.;
%end;
%local datamin datamax;

%if %length(&lowerknot.) NE 0 and %length(&lowerknot.) NE 16 %then %do;
   %PUT %str(ER)ROR: Lowernot should be given in a 16-characters Hex16. format. ; 
   %ABORT;  
%end;
%if %length(&upperknot.) NE 0 and %length(&upperknot.) NE 16 %then %do;
   %PUT %str(ER)ROR: Upperknot should be given in a 16-characters Hex16. format.; 
   %ABORT;  
%end;
%if &Nknots. >0 %then %do;
  %do m = 1 %to &Nknots.;
     %if %length(&&knot&m.) NE 0 and %length(&&knot&m.) NE 16 %then %do;
        %PUT %str(ER)ROR: Knot &m. should be given in a 16-characters Hex16. format.; 
        %ABORT;  
     %end;
  %end;
%end;


 %if %upcase(&indata.) NE _INSIDE %then %do;

 proc sql noprint;
   select put(min(&invariable.),hex16.), put(max(&invariable.),hex16.) into :datamin, :datamax
   from &indata. ;
 quit;

 %if &lowerknot. = %then %do; %let lowerknot = &datamin.; %end;
 %if &upperknot. = %then %do; %let upperknot = &datamax.; %end;

 %put Boundary knots are:  %sysfunc(inputn(&lowerknot.,hex16.)) (&lowerknot.) and %sysfunc(inputn(&upperknot.,hex16.)) (&upperknot.);


  data &outdata.;
    set &indata.;
 %end;

 %if &Nknots. > 0 %then %do; array knot[&Nknots.] _temporary_;  %end;
 array eknot[&dim.]   _temporary_ ;
 array aij[&dim.,&order.] _temporary_ ;
 array Bij[&dim.,&order.] _temporary_ ; *Note that ti = eknot(i+1) and ti+1=eknot(i+2);
 array &outarray.[&dimout.,%eval(&nparam.)] &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 

 %if &if. NE %then if &if then do;;
 %if &Nknots. >0 %then %do;
 %do m = 1 %to &Nknots.;
    knot[&m.]= input("&&knot&m.",hex16.);
 %end;
 %end;;
 
 do l = 1 to &order.+1; eknot[l]=input("&lowerknot.",hex16.); end;
 %if &Nknots. >0 %then %do; do l = &order.+2 to &order.+2+&Nknots.-1; eknot[l]=knot[l-&order.-1]; end; %end;
 do l = &order.+2+&Nknots. to &dim.; eknot[l]=input("&upperknot.",hex16.); end;

 do i = 0 to &dim.-1;
 do j=0 to &order.-1;
   if i+j < &dim.  and eknot[i+1] NE eknot[i+j+1] then aij[i+1,j+1]=(&invariable. - eknot[i+1])/(eknot[i+j+1]-eknot[i+1]);
   else aij[i+1,j+1]=0;
  end;
 end;
 do i = &dim.-1 to 0 by -1;
  do j=0 to &order.-1;
  if eknot[1] = &invariable. and j=0  then Bij[2,&order.]=1;
  if      j=0 and i<&dim.-1  and eknot[i+1]<&invariable.<=eknot[i+2] then Bij[i+1,j+1]=1; 
  else if j=0 then Bij[i+1,j+1]=0; 
  else if j>0 and i < &dim.-1 then Bij[i+1,j+1]=aij[i+1,j+1]*Bij[i+1,j]+(1-aij[i+2,j+1])*Bij[i+2,j]; 
  else Bij[i+1,j+1]=.;
  end;
 end;
 
 do k=0 to &nparam.-1;
  &outarray.[&posout.,k+1]=Bij[k+2, &order.];   
 end;

 if &invariable. >= input("&upperknot.",hex16.) then &outarray.[&posout.,&nparam.] = 1;
 if &invariable. <= input("&lowerknot.",hex16.) then &outarray.[&posout.,1] = 1;
 %if &if. NE %then end;;
drop  i j k l ;
 %if %upcase(&indata.) NE _INSIDE %then %do;
run;
 %end;
%if &incderiv. = YES and %upcase(&indata.) NE _INSIDE  %then %do;

%bspline(indata=&outdata., outdata=&outdata., invariable=&invariable.-&dif., outarray=delta_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., nparam_var=&nparam_var., if=&if.);

%bspline(indata=&outdata., outdata=&outdata., invariable=&invariable.+&dif., outarray=delta2_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., nparam_var=&nparam_var., if=&if.);

%if &incderiv2 = YES %then %do;
%bspline(indata=&outdata., outdata=&outdata., invariable=&invariable.-2*&dif., outarray=tdelta_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., nparam_var=&nparam_var., if=&if.);

%bspline(indata=&outdata., outdata=&outdata., invariable=&invariable.+2*&dif., outarray=tdelta2_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., nparam_var=&nparam_var., if=&if.);
%end;

data &outdata.;
  set &outdata.;
   array &outarray.[&dimout.,%eval(&nparam.)]  &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
   array delta_&outarray.[&dimout.,%eval(&nparam.)]    delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array delta2_&outarray.[&dimout.,%eval(&nparam.)]  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array drv_&outarray.[&dimout.,%eval(&nparam.)] drv_&outarray.0 - drv_&outarray.%eval(&dimout.*&nparam.-1) ; 

%if &incderiv2 = YES %then %do;
   array tdelta_&outarray.[&dimout.,%eval(&nparam.)]    tdelta_&outarray.0 -   tdelta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array tdelta2_&outarray.[&dimout.,%eval(&nparam.)]   tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array secdrv_&outarray.[&dimout.,%eval(&nparam.)] secdrv_&outarray.0 - secdrv_&outarray.%eval(&dimout.*&nparam.-1) ; 
%end;

  if &invariable.>input("&datamin.",hex16.) and &invariable.<input("&datamax.",hex16.) then do;
     do m = 1 to &nparam.;
	 
      drv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m] - delta_&outarray.[&posout.,m])/(2*&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m]   - 2*&outarray.[&posout.,m]  + delta_&outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;
     else if &invariable.=input("&datamin.",hex16.) then do;
     do m = 1 to &nparam.;
	 
      drv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m] - &outarray.[&posout.,m])/(&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(tdelta2_&outarray.[&posout.,m] -2*delta2_&outarray.[&posout.,m] + &outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;
	 else if &invariable.=input("&datamax.",hex16.) then do;
     do m = 1 to &nparam.;
	 
      drv_&outarray.[&posout.,m]=(&outarray.[&posout.,m]-delta_&outarray.[&posout.,m] )/(&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(&outarray.[&posout.,m] - 2*delta_&outarray.[&posout.,m] + tdelta_&outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;

 drop  m delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1)  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1);
 %if &incderiv2 = YES %then drop  tdelta_&outarray.0 - tdelta_&outarray.%eval(&dimout.*&nparam.-1)  tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1);;
run;
%end;

%mend bspline;
