**************************************************************************;
** Macro name      : kronrodrule15p                                     **;
** Version number  : 1.01                                               **;
** Authors         : Alberto Garcia-Hernandez                           **;
**                   Dimitris Rizopoulos                                **;
** Date            : 15-Dec-2014                                        **;
** SAS version     : 9.0 or higher                                      **;
** Purpose         : Create the 15 Kronrod Rule nodes required to       **;
**                   calculate the Integral from 0 to T of a function   **;
** Input Macro     : None                                               **;
** Input Data      : A variable with the point T                        **;
** Input Parameters: invariable  = Variable that stores T               **;
**                   outvariable = Array to store the 15 nodes          **;
**                                 and the 15 weights                   **;
**                   tempor     = 1 for xgk be _temporary_, 0 otherwise **;
**                                use tempor=0 when macro used within  **;
**                                proc nlmixed since  _temporary_ is    **;
**                                not allowed in nlmixed                **;
**                                                                      **;
** Output          : A variable/array with the 15 nodes of Kronrod Rule **;
**                   A variable/array with the 15 weights               **;
**                                                                      **;
** Purpose         : Calculate the Integral from 0 to T of a function   **;
**                                                                      **;
**                        MACRO HISTORY                                 **;
** Version     Date        Changes                                      **;
** 1.01        15/12/2014  First Production Version                     **;
**                                                                      **;
** Copyright (C) 2014      Alberto Garcia Hernandez                     **;
**                         Dimitris Rizopoulos                          **;
** This program is free software: you can redistribute it and/or modify **;
** it under the terms of the GNU General Public License as published by **;
** the Free Software Foundation, either version 3 of the License, or    **;
** (at your option) any later version.                                  **;
**                                                                      **;
** This program is distributed in the hope that it will be useful,      **;
** but WITHOUT ANY WARRANTY, without even the implied warranty of       **;
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        **;
** GNU General Public License for more details.                         **;
**                                                                      **;
** You should have received a copy of the GNU General Public License    **;
** along with this program.  If not, see <http://www.gnu.org/licenses/  **;
** Everyone is permitted to copy and distribute verbatim copies         **;
** of this program, but changing it is not allowed.                     **;
**                                                                      **;
** Contact email: alberto.garciahernandez@astellas.com                  **;;
**************************************************************************;

%macro kronrodrule15p(from=0, to=, nodes=, weights=, tempor=);
/*********************************************************************************************/
  /*  nodes of 15 point Kronrod rule  xgk 2,4,6,8,10,12 & 14 are also nodes of 7 point         */
   array xgk[8] %if &tempor. %then %do; _temporary_ %end;;   
   xgk[1]=0.991455371120812639206854697526329;
   xgk[2]=0.949107912342758524526189684047851;
   xgk[3]=0.864864423359769072789712788640926; 
   xgk[4]=0.741531185599394439863864773280788;
   xgk[5]=0.586087235467691130294144838258730; 
   xgk[6]=0.405845151377397166906606412076961;
   xgk[7]=0.207784955007898467600689403773245;
   xgk[8]=0;

   array &nodes.[15] ;

   half = 0.5*(&to.-&from.);                              *  half length of the interval;
   %do r = 1 %to 8;
	 &nodes.[&r.] = (&to.+&from.)*0.5 - half*xgk[&r.];
   %end;
   %do r = 9 %to 15;
     &nodes.[&r.] = (&to.+&from.)*0.5 + half*xgk[16-&r.];
   %end;
   drop half;


  array &weights.[15] ; 

   *  weights of 15 point Kronrod rule ;
if &to. ne . and &from. ne . then do;

       &weights.[1]=0.022935322010529224963732008058970;
       &weights.[2]=0.063092092629978553290700663189204;
       &weights.[3]=0.104790010322250183839876322541518;
       &weights.[4]=0.140653259715525918745189590510238;
       &weights.[5]=0.169004726639267902826583426598550;
       &weights.[6]=0.190350578064785409913256402421014;
       &weights.[7]=0.204432940075298892414161999234649;
       &weights.[8]=0.209482141084727828012999174891714;

   %do r = 9 %to 15;
     &weights.[&r.] = &weights.[16-&r.];
   %end;
end;
%mend kronrodrule15p;

