**************************************************************************;
** Macro name      : spline                                             **;
** Version number  : 1.02                                               **;
** Author          : Alberto Garcia-Hernandez                           **;
**                   Dimitris Rizopoulos                                **;
** Date            : 23-Jan-2015                                        **;
** SAS version     : 9.0 or higher                                      **;
** Purpose         : Calculate Spline Basis functions                   **;
** Input Data      : A variable with the point T                        **;
** Input Parameters: indata       = Input Dataset                       **;
**                   outdata      = Output Dataset                      **;
**                   invariable   = Variable that stores T              **;
**                   outarray     = Array/Matrix to store the basis     **;
**                   dimout = outarray  is a matrix of dimension        **;
**                            Dimeout x Number of basis functions       **; 
**                            use dimout=1 for the basis of variable T  **;
**                            use dimout>1 when bases are required for a**;
**                            vector of times                           **;  
**                   posout = position of the basis in the output matrix**;
**                   lowerknot= lower boundary knot in hex16. format    **;
**                   upperknot = upper boundary knot  in hex16. format  **;
**                   Nknots    = number of internal knots (max=3)       **;
**                   knot1, knot2, knot3 = internal knots value         **;
**                               in hex16. format     				    **;
**                   incderiv = YES if derivatives of the functions are **;
**                             required                                 **;
**                   incderiv2= YES if 2nd derivatives of the functions **;
**                             required                                 **;
**                   dif      = difference used to calculate derivatives**;
**                   if       = subset of records to apply the bases    **;
** Output          : Variable/Matrix with the spline Basis functions    **;
**                                                                      **;
**                        MACRO HISTORY                                 **;
** Version     Date        Changes                                      **;
** 1.01        15/12/2014  First Production Version                     **;
** 1.02        23/02/2015  New checks added for input knots not in      **;
**                         16-characters Hex16. format.                 **;
**                                                                      **;
** Copyright (C) 2014      Alberto Garcia Hernandez                     **;
**                         Dimitris Rizopoulos                          **;
** This program is free software: you can redistribute it and/or modify **;
** it under the terms of the GNU General Public License as published by **;
** the Free Software Foundation, either version 3 of the License, or    **;
** (at your option) any later version.                                  **;
**                                                                      **;
** This program is distributed in the hope that it will be useful,      **;
** but WITHOUT ANY WARRANTY, without even the implied warranty of       **;
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        **;
** GNU General Public License for more details.                         **;
**                                                                      **;
** You should have received a copy of the GNU General Public License    **;
** along with this program.  If not, see <http://www.gnu.org/licenses/  **;
** Everyone is permitted to copy and distribute verbatim copies         **;
** of this program, but changing it is not allowed.                     **;
**                                                                      **;
** Contact email: alberto.garciahernandez@astellas.com                  **;
**************************************************************************;

%macro spline(indata=, outdata=, invariable=, degree=3, lowerknot=, upperknot=, outarray=, dimout=1, posout=1, Nknots = 0,
               knot1=, knot2=, knot3=, knot4=, knot5=, knot6=, knot7=, knot8=, knot9=, 
               incderiv=NO, incderiv2=NO, normalize=YES, dif=0.001, nparam_var=longi_nparam, if=);

%let nparam=%eval(&degree. + 1 + &Nknots.);     *number of parameters;

%if &nparam_var. NE %then %do;
   %let &nparam_var. = &nparam.;
%end;

%local datamin datamax;

%if &Nknots. >0 %then %do;
  %do m = 1 %to &Nknots.;
     %if %length(&&knot&m.) NE 0 and %length(&&knot&m.) NE 16 %then %do;
        %PUT %str(ER)ROR: Knot &m. should be given in a 16-characters Hex16. format.; 
        %ABORT;  
     %end;
  %end;
%end;

 %if %upcase(&indata.) NE _INSIDE %then %do;
  proc sql noprint;
   select put(min(&invariable.),hex16.), put(max(&invariable.),hex16.) into :datamin, :datamax
   from &indata. ;
 quit;

 %if &lowerknot. = %then %do; %let lowerknot = &datamin.; %end;
 %if &upperknot. = %then %do; %let upperknot = &datamax.; %end;


  data &outdata.;
    set &indata.;
 %end;

 array &outarray.[&dimout.,%eval(&nparam.)] &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
 %if &if. NE %then if &if then do;;
 if &invariable. ne . then &outarray.[&posout.,1] = 1;
 %do d = 1 %to  &degree.;
 &outarray.[&posout.,%eval(&d.+1)] = (&invariable.)**&d. %if &normalize. =  YES %then / ((input("&upperknot.",hex16.)-input("&lowerknot.",hex16.))**&d.);; 
 %end;
 %if &Nknots. >0 %then %do;
     %do m = 1 %to &Nknots.;
	  %put Knot &m. is :  %sysfunc(inputn(&&knot&m.,hex16.)) ;
	         &outarray.[&posout.,%eval(&m.+&degree.+1)] = ((&invariable.)>input("&&knot&m.",hex16.))*
                                                          (&invariable.-input("&&knot&m.",hex16.))**&degree. %if &normalize. =  YES %then 
                                                          /((input("&upperknot.",hex16.)-input("&lowerknot.",hex16.))**&degree.);; 
      %end;
   %end;
%if &if. NE %then end;;

 %if %upcase(&indata.) NE _INSIDE %then %do;
run;
 %end;


%if &incderiv. = YES and %upcase(&indata.) NE _INSIDE %then %do;

%spline(indata=&outdata., outdata=&outdata., invariable=&invariable.-&dif., outarray=delta_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., normalize=&normalize., dif=&dif., if=&if., nparam_var=&nparam_var.);

%spline(indata=&outdata., outdata=&outdata., invariable=&invariable.+&dif., outarray=delta2_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., normalize=&normalize., dif=&dif., if=&if., nparam_var=&nparam_var.);

%if &incderiv2 = YES %then %do;
%spline(indata=&outdata., outdata=&outdata., invariable=&invariable.-2*&dif., outarray=tdelta_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., normalize=&normalize., dif=&dif., if=&if., nparam_var=&nparam_var.);

%spline(indata=&outdata., outdata=&outdata., invariable=&invariable.+2*&dif., outarray=tdelta2_&outarray., posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., degree=&degree., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., normalize=&normalize., dif=&dif., if=&if., nparam_var=&nparam_var.);
%end;

data &outdata.;
  set &outdata.;
   array &outarray.[&dimout.,%eval(&nparam.)]  &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
   array delta_&outarray.[&dimout.,%eval(&nparam.)]    delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array delta2_&outarray.[&dimout.,%eval(&nparam.)]  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array drv_&outarray.[&dimout.,%eval(&nparam.)] drv_&outarray.0 - drv_&outarray.%eval(&dimout.*&nparam.-1) ; 

%if &incderiv2 = YES %then %do;
   array tdelta_&outarray.[&dimout.,%eval(&nparam.)]    tdelta_&outarray.0 -   tdelta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array tdelta2_&outarray.[&dimout.,%eval(&nparam.)]   tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array secdrv_&outarray.[&dimout.,%eval(&nparam.)] secdrv_&outarray.0 - secdrv_&outarray.%eval(&dimout.*&nparam.-1) ; 
%end;

data &outdata.;
  set &outdata.;
   array &outarray.[&dimout.,%eval(&nparam.)]  &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
   array delta_&outarray.[&dimout.,%eval(&nparam.)]    delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array delta2_&outarray.[&dimout.,%eval(&nparam.)]  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array drv_&outarray.[&dimout.,%eval(&nparam.)] drv_&outarray.0 - drv_&outarray.%eval(&dimout.*&nparam.-1) ; 

%if &incderiv2 = YES %then %do;
   array tdelta_&outarray.[&dimout.,%eval(&nparam.)]    tdelta_&outarray.0 -   tdelta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array tdelta2_&outarray.[&dimout.,%eval(&nparam.)]   tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array secdrv_&outarray.[&dimout.,%eval(&nparam.)] secdrv_&outarray.0 - secdrv_&outarray.%eval(&dimout.*&nparam.-1) ; 
%end;

  if &invariable.>input("&datamin.",hex16.) and &invariable.<input("&datamax.",hex16.) then do;
     do m = 1 to &nparam.;
	 
      drv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m] - delta_&outarray.[&posout.,m])/(2*&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m]   - 2*&outarray.[&posout.,m]  + delta_&outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;
     else if &invariable.=input("&datamin.",hex16.) then do;
     do m = 1 to &nparam.;
	
      drv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m] - &outarray.[&posout.,m])/(&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(tdelta2_&outarray.[&posout.,m] -2*delta2_&outarray.[&posout.,m] + &outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;
	 else if &invariable.=input("&datamax.",hex16.) then do;
     do m = 1 to &nparam.;
	 
      drv_&outarray.[&posout.,m]=(&outarray.[&posout.,m]-delta_&outarray.[&posout.,m] )/(&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(&outarray.[&posout.,m] - 2*delta_&outarray.[&posout.,m] + tdelta_&outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;


 drop  m delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1)  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1);
 %if &incderiv2 = YES %then drop  tdelta_&outarray.0 - tdelta_&outarray.%eval(&dimout.*&nparam.-1)  tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1);;

run;
%end;


%mend spline;
